# -*- coding: utf-8 -*-

from django.http import Http404, HttpResponseNotAllowed
from djclsview import View

__all__ = ['Resource']


class Resource(View):

    def __init__(self, request, *args, **params):
        self.request = request
        self.args = args
        self.params = params
        self._called_yet = False

    def __call__(self):
        """Dispatch to an action based on HTTP method + URL."""

        # The problem with defining a resource as a callable is that a
        # reference to `self` from a Django template (in v1.3) will attempt to
        # call the resource again.
        if self._called_yet:
            return self
        self._called_yet = True

        method = self.request.POST.get('_method', self.request.method).upper()
        try:
            method_action_map = self.params.pop('methods')
        except KeyError:
            raise ValueError("Expected 'methods' dict in view kwargs")
        return self._route(method, method_action_map)()

    def _route(self, method, method_action_map):

        """
        Resolve an HTTP method and an HTTP method -> action mapping to a view.

        There are two sources for the list of 'defined methods' on a given URL:
        the HTTP method -> action map passed in to this method, and the actions
        which are defined on this `Resource` class. If the intersection of
        these two lists is empty--to wit, no methods are defined for the
        current URL--return a stub 404 view. Otherwise, if an HTTP method is
        sent which is not in *both* these lists, return a 405 'Not Allowed'
        view (which will contain the list of accepted methods at this URL).

        If the HTTP method sent is in the method -> action map, and the mapped
        action is defined on this `Resource`, return that action (which will be
        a callable `BoundAction` instance).
        """

        allowed_methods = self._allowed_methods(method_action_map)
        if method not in allowed_methods:
            # If *no* methods are defined for this URL, return a 404.
            if not allowed_methods:
                return not_found
            return lambda: HttpResponseNotAllowed(allowed_methods)

        action_name = method_action_map[method]
        return getattr(self, action_name)

    def _allowed_methods(self, method_action_map):
        allowed_methods = []
        for meth, action_name in method_action_map.items():
            if hasattr(self, action_name):
                allowed_methods.append(meth)
        return allowed_methods

    def _format(self):
        """Return a mimetype shortcode, in case there's no Accept header."""

        if self.params.get('format'):
            return self.params['format'].lstrip('.')
        return self.request.GET.get('format')


def not_found():
    """Stub function to raise `django.http.Http404`."""

    raise Http404



from django.conf.urls import defaults


class URLRouter(object):

    """
    Responsible for generating include()-able URLconfs for resources.

    Accepts only one argument on instantiation, `style`. This should be a
    callable which accepts an action parameter, a routing mode and an ID regex,
    and returns a regular expression with a route for that action. You should
    only need to use the styles already defined in `dagny.urls.styles`.
    """

    def __init__(self, style):
        self.style = style

    def _make_patterns(self, resource_name, id, name, actions, urls):

        """
        Construct an `include()` with all the URLs for a resource.

        :param resource_name:
            The full path to the resource (e.g.  `myapp.resources.User`).
        :param id:
            The ID parameter, either as a regex fragment or a `(name, regex)`
            pair, which will normally be translated by the URL style to a named
            group in the URL.
        :param name:
            The name for this resource, which will be used to generate named
            URL patterns (e.g. if name is 'User', URLs will be 'User#index',
            'User#show' etc.). If `None`, defaults to `resource_name`.
        :param actions:
            The actions to generate (named) routes for. If `None`, a route and
            a name will be generated for every one defined by the chosen URL
            style.
        :param urls:
            A list of the URLs to define patterns for. Must be made up only of
            'member', 'collection', 'new', 'edit', 'singleton' and
            'singleton_edit'.
        """

        if actions is not None:
            actions = set(actions)
        if name is None:
            name = resource_name

        urlpatterns = []
        for url in urls:
            # URLStyle.__call__(url_name, id_pattern)
            #     => (url_pattern, {method: action, ...})
            pattern, methods = self.style(url, id)
            # Filter methods dict to only contain the selected actions.
            methods = dict(
                (method, action) for method, action in methods.iteritems()
                if (actions is None) or (action in actions))
            # Add named url patterns, one per action. Note that we will have
            # duplicate URLs in some cases, but this is so that
            # `{% url User#show %}` can be distinguished from
            # `{% url User#update %}` when it makes sense.
            for action in methods.itervalues():
                urlpatterns.append(defaults.url(pattern, resource_name,
                                                kwargs={'methods': methods},
                                                name=("%s#%s" % (name, action))))
        return defaults.include(defaults.patterns('', *urlpatterns))

    def resources(self, resource_name, id=r'\d+', actions=None, name=None):
        return self._make_patterns(resource_name, id, name, actions,
                                   ['collection', 'new', 'member', 'edit'])

    def resource(self, resource_name, actions=None, name=None):
        return self._make_patterns(resource_name, '', name, actions,
                                   ['singleton', 'new', 'singleton_edit'])



class URLStyle(object):

    """
    Generic class for defining resource URL styles.

    `URLStyle` can be used to create callables which will work for the
    interface defined in `dagny.urls.router.URLRouter`. Subclass and override
    the `collection()`, `new()`, `member()`, `edit()`, `singleton()` and
    `singleton_edit()` methods to customize your URLs. You can use one of the
    several defined styles in this module as a template.
    """

    METHODS = {
        'collection': {
            'GET': 'index',
            'POST': 'create'
        },
        'member': {
            'GET': 'show',
            'POST': 'update',
            'PUT': 'update',
            'DELETE': 'destroy'
        },
        'new': {'GET': 'new'},
        'edit': {'GET': 'edit'},
        'singleton': {
            'GET': 'show',
            'POST': 'update',
            'PUT': 'update',
            'DELETE': 'destroy'
        },
        'singleton_edit': {'GET': 'edit'},
    }

    def __call__(self, url, id_param):
        id_regex = self._get_id_regex(id_param)

        if url in ('member', 'edit'):
            return getattr(self, url)(id_regex), self.METHODS[url]
        return getattr(self, url)(), self.METHODS[url]

    def _get_id_regex(self, id_param):

        """
        Resolve `(name, regex)` => `'?P<name>regex'`.

        Since the style methods should add parentheses around the ID regex
        fragment, the output for named ID parameters is not surrounded by
        parentheses itself.
        """

        if isinstance(id_param, basestring):
            return id_param
        elif isinstance(id_param, tuple):
            if len(id_param) != 2:
                raise ValueError("id param must be a (name, regex) pair")
            name, regex = id_param
            return '?P<%s>%s' % (name, regex)
        raise TypeError('id param must be a string or (name, regex) pair, '
                        'not %r' % (type(id_param),))

    # Publicly-overrideable methods for customizing style behaviour.

    def collection(self):
        raise NotImplementedError

    def new(self):
        raise NotImplementedError

    def member(self, id_regex):
        raise NotImplementedError

    def edit(self, id_regex):
        raise NotImplementedError

    def singleton(self):
        raise NotImplementedError

    def singleton_edit(self):
        raise NotImplementedError


class DjangoURLStyle(URLStyle):

    """
    Standard Django-style URLs.

       URL            | action | args   | kwargs
       ---------------+--------+--------+--------
       /posts/        | index  | ()     | {}
       /posts/new/    | new    | ()     | {}
       /posts/1/      | show   | ('1',) | {}
       /posts/1/edit/ | edit   | ('1',) | {}
    """

    def collection(self):
        return r'^$'

    def new(self):
        return r'^new/$'

    def member(self, id_regex):
        return r'^(%s)/$' % (id_regex,)

    def edit(self, id_regex):
        return r'^(%s)/edit/$' % (id_regex,)

    def singleton(self):
        return r'^$'

    def singleton_edit(self):
        return r'^edit/$'


class AtomPubURLStyle(URLStyle):

    """
    Atom Publishing Protocol-style URLs.

    The main difference between this and Django-style URLs is the lack of
    trailing slashes on leaf nodes.

        URL           | action | args   | kwargs
        --------------+--------+--------+--------
        /posts/       | index  | ()     | {}
        /posts/new    | new    | ()     | {}
        /posts/1      | show   | ('1',) | {}
        /posts/1/edit | edit   | ('1',) | {}
    """

    def collection(self):
        return r'^$'

    def new(self):
        return r'^new$'

    def member(self, id_regex):
        return r'^(%s)$' % (id_regex,)

    def edit(self, id_regex):
        return r'^(%s)/edit$' % (id_regex,)

    def singleton(self):
        return r'^$'

    def singleton_edit(self):
        return r'^edit$'


class RailsURLStyle(URLStyle):

    r"""
    Ruby on Rails-style URLs.

    This URL style is quite advanced; it will also capture format extensions
    and pass them through as a kwarg. As with `AtomPubURLStyle`, trailing
    slashes are not mandatory on leaf nodes.

        URL           | action | args | kwargs
        --------------+--------+------+------------------------------
        /posts        | index  | ()   | {}
        /posts.json   | index  | ()   | {'format': 'json'}
        /posts/       | index  | ()   | {}
        /posts/new    | new    | ()   | {}
        /posts/1      | show   | ()   | {'id': '1'}
        /posts/1.json | show   | ()   | {'id': '1', 'format': 'json'}
        /posts/1/     | show   | ()   | {'id': '1'}
        /posts/1/edit | edit   | ()   | {'id': '1'}

    **Note**: due to limitations of the URLconf system, your IDs/slugs have to
    come in as named parameters. By default, the parameter will be called `id`,
    but you can select a different one using the `id` keyword argument to the
    URL helpers:

        urlpatterns = patterns('',
            (r'posts', resources('myapp.resources.Post', name='Post',
                                 id=('slug', r'[\w\-]+'))),
        )

    Another caveat: do not terminate your inclusion regex with a slash, or the
    format extension on the resource index won't work.

    You can customize the format extension regex (and hence the kwarg name) by
    subclassing and overriding the `FORMAT_EXTENSION_RE` attribute, e.g.:

        class MyRailsURLStyle(RailsURLStyle):
            FORMAT_EXTENSION_RE = r'(?P<accept>[A-Za-z0-9]+)'
    """

    FORMAT_EXTENSION_RE = r'(?P<format>\.\w[\w\-\.]*)'

    def _get_id_regex(self, id_param):
        """Co-erce *all* IDs to named parameters, defaulting to `'id'`."""

        if isinstance(id_param, basestring) and not id_param.startswith('?P<'):
            return super(RailsURLStyle, self)._get_id_regex(('id', id_param))
        return super(RailsURLStyle, self)._get_id_regex(id_param)

    def collection(self):
        return r'^%s?/?$' % (self.FORMAT_EXTENSION_RE,)

    def new(self):
        return r'^/new/?$'

    def member(self, id_regex):
        return r'^/(%s)%s?/?$' % (id_regex, self.FORMAT_EXTENSION_RE)

    def edit(self, id_regex):
        return r'^/(%s)/edit/?$' % (id_regex,)

    def singleton(self):
        return r'^%s?/?$' % (self.FORMAT_EXTENSION_RE,)

    def singleton_edit(self):
        return r'^/edit/?$'
